	;; 
	;; Copyright (c) 2005 Axis Communications AB
	;; 
	;; Fast AES for crisv10. Meassured to be up to 25% faster than
	;; OpenSSL's 0.9.7g C-implementation compiled with gcc-cris R60.
	;;
	;; Loops are not unrolled to improve cache locality.
	;;
	
	;; T is the table to look up in
	;; i0 - i3 index variables (s0-s3 or t0-t3 depending on the state)
	;; d0 - d3 destination variable
	;; messes up r4-r6

	.macro	aes_tablelookup_1	T, i0, i1, i2, i3, d0, d1, d2, d3
	move.d	\i0, $r4
	lsrq	24, $r4	
	move.d	\i1, $r5
	lsrq	24, $r5

	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	24, $r4
	move.d	\i3, $r5
	lsrq	24, $r5
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm
		
	.macro	aes_tablelookup_2	T, i0, i1, i2, i3, d0, d1, d2, d3
	;; second table lookup
	move.d	\i0, $r4
	lsrq	14, $r4
	and.d	0x3fc, $r4
	move.d	\i1, $r5
	lsrq	14, $r5
	and.d	0x3fc, $r5

	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	14, $r4
	and.d	0x3fc, $r4
	
	move.d	\i3, $r5
	lsrq	14, $r5
	and.d	0x3fc, $r5
	
	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm

	.macro	aes_tablelookup_3	T, i0, i1, i2, i3, d0, d1, d2, d3
	;; second table lookup
	move.d	\i0, $r4
	lsrq	6, $r4
	and.d	0x3fc, $r4
	move.d	\i1, $r5
	lsrq	6, $r5
	and.d	0x3fc, $r5

	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	6, $r4
	and.d	0x3fc, $r4
	
	move.d	\i3, $r5
	lsrq	6, $r5
	and.d	0x3fc, $r5
	
	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm

	.macro	aes_tablelookup_4	T, i0, i1, i2, i3, d0, d1, d2, d3
	movu.b	\i0, $r4
	movu.b	\i1, $r5
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	movu.b	\i2, $r4
	movu.b	\i3, $r5	
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	xor	$r6, \d2
	.endm
	
	.macro	aes_tablelookup_4_full	T, i0, i1, i2, i3, d0, d1, d2, d3
	aes_tablelookup_4	\T, \i0, \i1, \i2, \i3, \d0, \d1, \d2, \d3
	xor	$r4, \d3
	.endm

	.macro	aes_tablelookup_final_1	T, mask, i0, i1, i2, i3, d0, d1, d2, d3
	move.d	\i0, $r4
	lsrq	24, $r4	
	move.d	\i1, $r5
	lsrq	24, $r5

	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	24, $r4
	move.d	\i3, $r5
	lsrq	24, $r5
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm
	
	.macro	aes_tablelookup_final_2	T, mask, i0, i1, i2, i3, d0, d1, d2, d3
	move.d	\i0, $r4
	lsrq	14, $r4
	and.d	0x3fc, $r4
	move.d	\i1, $r5
	lsrq	14, $r5
	and.d	0x3fc, $r5

	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	14, $r4
	and.d	0x3fc, $r4
	
	move.d	\i3, $r5
	lsrq	14, $r5
	and.d	0x3fc, $r5
	
	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm

	.macro	aes_tablelookup_final_3	T, mask, i0, i1, i2, i3, d0, d1, d2, d3
	move.d	\i0, $r4
	lsrq	6, $r4
	and.d	0x3fc, $r4
	move.d	\i1, $r5
	lsrq	6, $r5
	and.d	0x3fc, $r5

	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	move.d	\i2, $r4
	lsrq	6, $r4
	and.d	0x3fc, $r4
	
	move.d	\i3, $r5
	lsrq	6, $r5
	and.d	0x3fc, $r5
	
	move.d	[\T + $r4.b], $r6
	move.d	[\T + $r5.b], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm
	
	.macro	aes_tablelookup_final_4	T, mask, i0, i1, i2, i3, d0, d1, d2, d3
	movu.b	\i0, $r4
	movu.b	\i1, $r5
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d0
	xor	$r4, \d1
	
	movu.b	\i2, $r4
	movu.b	\i3, $r5	
	move.d	[\T + $r4.d], $r6
	move.d	[\T + $r5.d], $r4
	and.d	\mask, $r6
	and.d	\mask, $r4
	xor	$r6, \d2
	xor	$r4, \d3
	.endm
			
	;; void AES_encrypt(const unsigned char *in, unsigned char *out,
	;;                  const AES_KEY *key)
#define S_SAVEDREGS     0
#define S_OUT		S_SAVEDREGS + (9 * 4)
#define S_END		S_OUT + 4

	;; export the tables
	.text
	.global	_AES_encrypt
	.global	AES_encrypt
	.balign	32
_AES_encrypt:
AES_encrypt:
	subu.w	S_END, $sp
	movem	$r8, [$sp]

	move.d	$r11, [$sp + S_OUT]
aes_load_e_tables:	
	;; pic to load the encryption tables
	move.d	$pc, $r13
	add.d	Te0 - aes_load_e_tables - 2, $r13
	
	movem	[$r10], $r3	
	move.d	[$r12], $r7
	move.d	[$r12 + 4], $r6
	move.d	[$r12 + 8], $r5	
	move.d	[$r12 + 12], $r4
	addq	4*4, $r12
	
	;; big-endian
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	;; r3 - r0 now hold s0 - s3
	xor	$r4, $r0
	xor	$r5, $r1
	xor	$r6, $r2
	xor	$r7, $r3

	;; nr rounds
	move.d	[$r12 + 240 - 4*4], $r7 
	
	;; r0-r4  = s0-s3
	;; r4-r7  = are free
	;; r8-r11 = t0-t3
	;;
	;; The reverse order of the registers vs the variable names
	;; is not for fun, it is beacuse it enables the possibility
	;; to grab the key for both round halfs in one go (movem).
	;; 
	;; r7 is our loop-counter
	lsrq	1, $r7
aes_e_rounds:
	move.d	$r3, $r4
	lsrq	24, $r4	
	move.d	$r2, $r5
	lsrq	24, $r5
	move.d	$r1, $r6
	lsrq	24, $r6
	move.d	[$r13 + $r4.d], $r11
	move.d	$r0, $r4
	lsrq	24, $r4
	move.d	[$r13 + $r5.d], $r10
	move.d	[$r13 + $r6.d], $r9
	move.d	[$r13 + $r4.d], $r8
	
	addu.w	256*4, $r13		
	aes_tablelookup_2	$r13, $r2, $r1, $r0, $r3, $r11, $r10, $r9, $r8
	addu.w	256*4, $r13
	aes_tablelookup_3	$r13, $r1, $r0, $r3, $r2, $r11, $r10, $r9, $r8
	addu.w	256*4, $r13
	aes_tablelookup_4_full	$r13, $r0, $r3, $r2, $r1, $r11, $r10, $r9, $r8

	;; now get the key, and initialize s0-s3 in one go
	addq	4, $r12
	movem	[$r12+], $r6		
	xor	$r4, $r8
	move.d	[$r12 - 8*4], $r4
	xor	$r5, $r9
	xor	$r6, $r10
		
	subq	1, $r7
	beq	aes_e_lastround
	xor	$r4, $r11
	
	;; do second part of the rounds
	subu.w	256 * 4 * 3, $r13
		
	aes_tablelookup_1	$r13, $r11, $r10, $r9, $r8, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_2	$r13, $r10, $r9, $r8, $r11, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_3	$r13, $r9, $r8, $r11, $r10, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_4	$r13, $r8, $r11, $r10, $r9, $r3, $r2, $r1, $r0

	subu.w	256 * 4 * 3, $r13
	ba	aes_e_rounds
	xor	$r4, $r0	; see chomped aes_tablelookup_4 vs full
aes_e_lastround:
	;; apply last round, mapping in the internal state into the
	;; output buffer

	movu.b	0xff, $r12
	lslq	24, $r12
	addu.w	256*4, $r13		
		
	aes_tablelookup_final_1	$r13, $r12, $r11, $r10, $r9, $r8, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_2	$r13, $r12, $r10, $r9, $r8, $r11, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_3	$r13, $r12, $r9, $r8, $r11, $r10, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_4	$r13, $r12, $r8, $r11, $r10, $r9, $r3, $r2, $r1, $r0	

	move.d	[$sp + S_OUT], $r5
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	movem	$r3, [$r5]
		
	movem	[$sp], $r8
	addu.w	S_END, $sp
	ret
	nop

	;; export the tables
	.global	_Te0
	.global	_Te1
	.global	_Te2
	.global	_Te3
	.global	_Te4
	.global	_Td0
	.global	_Td1
	.global	_Td2
	.global	_Td3
	.global	_Td4
	.global	Te0
	.global	Te1
	.global	Te2
	.global	Te3
	.global	Te4
	.global	Td0
	.global	Td1
	.global	Td2
	.global	Td3
	.global	Td4

	;;
	;; The tables are put in the middle of encrypt/decrypt to
	;; make sure that no part of the code doing the encryption
	;; or decryption collides with cache-lines holding data from
	;; the tables. The tables are 5120*2 bytes but only the first
	;; half is used for encrypt (the second half for decrypt).
	;; This gives us about 3K on each side to avoid collisions.
	;; 
	;; The tables are consecutive to save registers in the round
	;; loops. Only one pointer to them is needed, $r13.
	;; 
	.balign	4
#include	"crisv10_aes_tables.S"
	
	.global	_AES_decrypt
	.global	AES_decrypt
_AES_decrypt:
AES_decrypt:
	subu.w	S_END, $sp
	movem	$r8, [$sp]

	move.d	$r11, [$sp + S_OUT]
aes_load_d_tables:
	;; pic to load the decryption tables
	move.d	$pc, $r13
	sub.d	aes_load_d_tables - Td0 + 2, $r13

	movem	[$r10], $r3	
	move.d	[$r12], $r7
	move.d	[$r12 + 4], $r6
	move.d	[$r12 + 8], $r5	
	move.d	[$r12 + 12], $r4
	addq	4*4, $r12
	
	;; big-endian
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	;; r3 - r0 now hold s0 - s3
	xor	$r4, $r0
	xor	$r5, $r1
	xor	$r6, $r2
	xor	$r7, $r3

	;; nr rounds
	move.d	[$r12 + 240 - 4*4], $r7 

	;; r0-r4  = s0-s3
	;; r4-r7  = are free
	;; r8-r11 = t0-t3
	;; r7 is our loop-counter
	lsrq	1, $r7	
aes_d_rounds:
	move.d	$r3, $r4
	lsrq	24, $r4	
	move.d	$r2, $r5
	lsrq	24, $r5
	move.d	$r1, $r6
	lsrq	24, $r6
	move.d	[$r13 + $r4.d], $r11
	move.d	$r0, $r4
	lsrq	24, $r4
	move.d	[$r13 + $r5.d], $r10
	move.d	[$r13 + $r6.d], $r9
	move.d	[$r13 + $r4.d], $r8

	addu.w	256*4, $r13		
	aes_tablelookup_2	$r13, $r0, $r3, $r2, $r1, $r11, $r10, $r9, $r8
	addu.w	256*4, $r13
	aes_tablelookup_3	$r13, $r1, $r0, $r3, $r2, $r11, $r10, $r9, $r8
	addu.w	256*4, $r13
	aes_tablelookup_4_full	$r13, $r2, $r1, $r0, $r3, $r11, $r10, $r9, $r8

	;; now get the key, and initialize s0-s3 in one go
	addq	4, $r12
	movem	[$r12+], $r6		
	xor	$r4, $r8
	move.d	[$r12 - 8*4], $r4
	xor	$r5, $r9
	xor	$r6, $r10
		
	subq	1, $r7
	beq	aes_d_lastround
	xor	$r4, $r11
	
	;; do second part of the rounds
	subu.w	256 * 4 * 3, $r13
		
	aes_tablelookup_1	$r13, $r11, $r10, $r9, $r8, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_2	$r13, $r8, $r11, $r10, $r9, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_3	$r13, $r9, $r8, $r11, $r10, $r3, $r2, $r1, $r0
	addu.w	256*4, $r13
	aes_tablelookup_4	$r13, $r10, $r9, $r8, $r11, $r3, $r2, $r1, $r0
	
	subu.w	256 * 4 * 3, $r13
	ba	aes_d_rounds
	xor	$r4, $r0	; see chomped aes_tablelookup_4 vs full
aes_d_lastround:
	;; apply last round, mapping in the internal state into the
	;; output buffer
	
	movu.b	0xff, $r12
	lslq	24, $r12
	addu.w	256*4, $r13		
		
	aes_tablelookup_final_1	$r13, $r12, $r11, $r10, $r9, $r8, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_2	$r13, $r12, $r8, $r11, $r10, $r9, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_3	$r13, $r12, $r9, $r8, $r11, $r10, $r3, $r2, $r1, $r0
	lsrq	8, $r12
	aes_tablelookup_final_4	$r13, $r12, $r10, $r9, $r8, $r11, $r3, $r2, $r1, $r0	

	move.d	[$sp + S_OUT], $r5
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	movem	$r3, [$r5]
		
	movem	[$sp], $r8
	addu.w	S_END, $sp
	ret
	nop
