	;; 
	;; Copyright (c) 2005 Axis Communications AB
	;; 
	;; Fast SHA-1 for crisv10. Meassured to be up to 70% faster than
	;; OpenSSL's 0.9.7g C-implementation compiled with gcc-cris R60.
	;;
	;; Simulator shows a decrease in executed cycles of approx 10%.
	;;
	;; Loops are not unrolled to improve cache locality.
	;;

	;; these are offsets into the stack to keep track of our local scoped
	;; auto variables
#define	S_SAVEDREGS	0
#define	S_W		S_SAVEDREGS + (9 * 4)
#define S_BLOCKLOOP	S_W + (80*4)
#define	S_END		S_BLOCKLOOP + 4

	;; the sha1 shift, clobbers both regs
	.macro sha1_shift bits,reg,dreg
	;; where is rol on cris?
	lslq	\bits,\reg
	lsrq	32 - \bits,\dreg
	or.d	\reg, \dreg
	.endm

	;; no args, soz
	.macro	sha1_compute_w
	;; b. For t = 16 to 79 let
	;; W(t) = S^1(W(t-3) XOR W(t-8) XOR W(t-14) XOR W(t-16)).
	move.d	[$r13 - 3*4], $r8	; r8 = W[t-3]	
	move.d	[$r13 - 8*4], $r9	; r9 = W[t-8]
	xor	$r9, $r8
	move.d	[$r13 - 14*4], $r9	; r9 = W[t-14]
	xor	$r9, $r8
	move.d	[$r13 - 16*4], $r9	; r9 = W[t-16]
	xor	$r9, $r8
	
	move.d	$r8, $r9
	sha1_shift 1,$r9,$r8
	
	;; write-back
	move.d	$r8, [$r13+]

	;; all loops add E + K[x]
	add.d	$r0, $r8	; + E
	add.d	$r5, $r8	; + K[x]
	.endm

	.macro	sha1_0_20_round
	;; TEMP = S^5(A) + f(t;B,C,D) + E + W[s] + K(t);
        move.d  $r1, $r0        ; E = D
        move.d  $r2, $r1        ; D = C

	move.d	$r3, $r9
	not	$r9
	and.d	$r0, $r9	; ~B & D
	and.d	$r3, $r2	;  B & C
	or.d	$r2, $r9
	add.d	$r9, $r8

	move.d	$r3, $r2
        sha1_shift 30, $r3, $r2  ; C = SHA1CircularShift(30,B);
        move.d  $r4, $r3        ; B = A

	;; r4 is now cloberable, take the opportunity to shift
	move.d	$r4, $r9
	sha1_shift 5,$r4,$r9
	add.d	$r9, $r8	; + SHA1CircularShift(5,A)		
	.endm
	
	;; void HASH_BLOCK_HOST_ORDER (SHA_CTX *c, const void *p, int num)
	.global	_sha1_block_asm_host_order
	.global	sha1_block_asm_host_order
	.balign	32
_sha1_block_asm_host_order:
sha1_block_asm_host_order:
	subu.w	S_END, $sp
	movem	$r8, [$sp]

        move.d  multiple_blocks_return - multiple_blocks_host_loop, $r9
        move.d  $r9, [$sp + S_BLOCKLOOP]

multiple_blocks_host_loop:			
	;; initialize w(0) - w(15) with the message, its in host order (LE).
	movem	[$r11+], $r7
	movem	$r7, [$sp + S_W]	
	movem	[$r11+], $r7
	movem	$r7, [$sp + S_W + 8 * 4]
	
	;; load A, B, C, D, E into r0 - r4
	movem	[$r10], $r4

	ba	sha1_common
	moveq	16, $r7		; prepare for 0_16 loop
	
	;; void HASH_BLOCK_DATA_ORDER (SHA_CTX *c, const void *p, int num)
	.global	_sha1_block_asm_data_order
	.global	sha1_block_asm_data_order
_sha1_block_asm_data_order:
sha1_block_asm_data_order:
	subu.w	S_END, $sp
	movem	$r8, [$sp]

        move.d  multiple_blocks_return - multiple_blocks_data_loop, $r9
        move.d  $r9, [$sp + S_BLOCKLOOP]

multiple_blocks_data_loop:		
	;; initialize w(0) - w(15) with the message, data order (BE).
	movem	[$r11+], $r7
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	swapwb	$r4
	swapwb	$r5
	swapwb	$r6
	swapwb	$r7	
	movem	$r7, [$sp + S_W]	
	movem	[$r11+], $r7
	swapwb	$r0
	swapwb	$r1
	swapwb	$r2
	swapwb	$r3
	swapwb	$r4
	swapwb	$r5
	swapwb	$r6
	swapwb	$r7	
	movem	$r7, [$sp + S_W + 8 * 4]	

	;; load A, B, C, D, E into r0 - r4
	movem	[$r10], $r4
	ba	sha1_common
	moveq	16, $r7		; prepare for 0_16 loop

        ;;
        ;; r0 - r4 holds A, B, C, D, E
        ;; r5 holds the current K[x] in use, depends on t
        ;; r8 used while computing and later holding W[t]
        ;; r13 holds W + t (pointer that is)
        ;;	
	.balign	32	
sha1_common:
	;; let r13 point at start of W
	move.d	$sp, $r13
	addq	S_W, $r13
	
	move.d	0x5A827999, $r5	; K[0]
sha1_0_16_loop:
	;; ((B & C) | ((~B) & D)) + E + W[t] + K[0];
	move.d	[$r13+], $r8	; r8 = W[t]
	add.d	$r5, $r8	; + K[0]
	add.d	$r0, $r8	; + E

	sha1_0_20_round		
        subq    1, $r7
        bgt     sha1_0_16_loop
        move.d  $r8, $r4       ; A = temp
	
	moveq	4, $r7		; prepare for 0_16 loop
sha1_16_20_loop:

	sha1_compute_w	
	sha1_0_20_round		

        subq    1, $r7
        bgt     sha1_16_20_loop
        move.d  $r8, $r4       ; A = temp
				
	move.d	0x6ED9EBA1, $r5	; load K[1] into r5
	moveq	20, $r7		; 20 * 4	
sha1_20_40_loop:
	sha1_compute_w	

	;; SHA1CircularShift(5,A) + (B ^ C ^ D) + E + W[t] + K[1];
	move.d	$r1, $r0	; E = D, r1 is now free

	;; B^D C^D == (B ^ C ^ D)
	xor	$r3, $r1
	xor	$r2, $r1
	add.d	$r1, $r8	; r8 is almost done
	
	move.d	$r2, $r1	; D = C
	move.d	$r3, $r2
	sha1_shift 30, $r3, $r2	; C = SHA1CircularShift(30,B);
	move.d	$r4, $r3	; B = A

	;; r4 is now cloberable, take the opportunity to shift
	move.d	$r4, $r9
	sha1_shift 5,$r4,$r9
	add.d	$r9, $r8	; + SHA1CircularShift(5,A)		
					
	subq	1, $r7
	bgt	sha1_20_40_loop
	move.d	$r8, $r4	; A = temp
	
	move.d	0x8F1BBCDC, $r5	; load K[2] into r5
	moveq	20, $r7		; prepare for 0_16 loop
sha1_40_60_loop:
	sha1_compute_w	

	;;((B & C) | (B & D) | (C & D)) + E + W[t] + K[2];
	move.d	$r1, $r0	; E = D

	move.d	$r1, $r9
	and.d	$r2, $r1	; C & D
	and.d	$r3, $r9	; B & D
	or.d	$r1, $r9	; (B & D) | (C & D)
	
	move.d	$r2, $r1	; D = C, $r2 is now free
	and.d	$r3, $r2	; B & C
	or.d	$r2, $r9	; final OR
	add.d	$r9, $r8	; temp is almost done now

	move.d	$r3, $r2
	sha1_shift 30, $r3, $r2	; C = SHA1CircularShift(30,B);
	move.d	$r4, $r3	; B = A
	
	;; r4 is now cloberable, take the opportunity to shift
	move.d	$r4, $r9
	sha1_shift 5,$r4,$r9
	add.d	$r9, $r8	; + SHA1CircularShift(5,A)		
	
	subq	1, $r7
	bgt	sha1_40_60_loop
	move.d	$r8, $r4	; A = temp
	
	move.d	0xCA62C1D6, $r5	; load K[3] into r5
	moveq	20, $r7		; prepare for 0_16 loop
sha1_60_80_loop:
	sha1_compute_w
	
	;; this is the same as 20_40 but with a different K - $r5
	
	;; SHA1CircularShift(5,A) + (B ^ C ^ D) + E + W[t] + K[1];
	move.d	$r1, $r0	; E = D, r1 is now free

	;; B^D C^D == (B ^ C ^ D)
	xor	$r3, $r1
	xor	$r2, $r1
	add.d	$r1, $r8	; r8 is almost done
	
	move.d	$r2, $r1	; D = C
	move.d	$r3, $r2
	sha1_shift 30, $r3, $r2	; C = SHA1CircularShift(30,B);
	move.d	$r4, $r3	; B = A

	;; r4 is now cloberable, take the opportunity to shift
	move.d	$r4, $r9
	sha1_shift 5,$r4,$r9
	add.d	$r9, $r8	; + SHA1CircularShift(5,A)		
					
	subq	1, $r7
	bgt	sha1_60_80_loop
	move.d	$r8, $r4	; A = temp
	
	;; load A, B, C, D, E into r0 - r4, r10 will be clobbered
	add.d	[$r10], $r4
	add.d	[$r10 + 4], $r3
	add.d	[$r10 + 8], $r2
	add.d	[$r10 + 12], $r1
	add.d	[$r10 + 16], $r0
	
	movem	$r4,[$r10]
	;; multiple blocks loop for openssl
	subq	1, $r12
	ble	multiple_blocks_return
	moveq	16, $r7		; prepare for 0_16 loop
	subu.w	[$sp + S_BLOCKLOOP], $pc
multiple_blocks_return:
		
	movem	[$sp], $r8
	addu.w	S_END, $sp
	ret
	nop
